/*
  This is polynomials.h
  
  Coxeter version 3.0 Copyright (C) 2002 Fokko du Cloux
  See file main.cpp for full copyright notice
*/

#ifndef POLYNOMIALS_H  /* include guard */
#define POLYNOMIALS_H

#include "globals.h"
#include <limits.h>

namespace polynomials {
  using namespace globals;
};

/******** type declarations **************************************************/

namespace polynomials {
  typedef Ulong Degree;
  typedef long SDegree;
  class Monomial;
  template <class T> class Polynomial;
  template <class T> class LaurentPolynomial;
};

/******** constants **********************************************************/

namespace polynomials {
  const Degree undef_degree = ~0;
  const Degree DEGREE_MAX = ULONG_MAX-1;
  const SDegree SDEGREE_MAX = LONG_MAX;
  const SDegree SDEGREE_MIN = LONG_MIN+1;
  const SDegree undef_valuation = LONG_MIN;
};

/******** function definitions ***********************************************/

#include "io.h"

namespace polynomials {
  using namespace io;
};

namespace polynomials {
  template <class T> 
    bool operator== (const Polynomial<T>& p, const Polynomial<T>& q);
  template <class T> 
  bool operator!= (const Polynomial<T>& p, const Polynomial<T>& q);
  template <class T> 
  bool operator<= (const Polynomial<T>& p, const Polynomial<T>& q);
  template <class T> 
  bool operator>= (const Polynomial<T>& p, const Polynomial<T>& q);
  template <class T> 
  bool operator< (const Polynomial<T>& p, const Polynomial<T>& q);
  template <class T> 
  bool operator> (const Polynomial<T>& p, const Polynomial<T>& q);
  template <class T>
  String& append(String& str, const Polynomial<T> &p, const char *x);
  template <class T>
  String& append(String& str, const Polynomial<T>& p, const Degree& d,
	     const long& m, const char *x);
  template <class T>
  String& append(String& str, const Polynomial<T>& p, const Degree& d,
	     const long& m, const char *x,GAP);
  template <class T>
  String& append(String& str, const Polynomial<T>& p, const Degree& d,
	     const long& m, const char *x,Terse);
  template <class T>
  void print(FILE* file, const Polynomial<T>& p, const char *x);
  template <class T>
  void print(FILE* file, const LaurentPolynomial<T>& p, const char *x);
  template <class T>
  void print(FILE* file, const Polynomial<T>& p, const Degree& d,
	     const long& m, const char *x);
  template <class T>
  void print(FILE* file, const Polynomial<T>& p, const Degree& d,
	     const long& m, const char *x,GAP);
  template <class T>
  void print(FILE* file, const Polynomial<T>& p, const Degree& d,
	     const long& m, const char *x,Terse);
  template <class T>
  SDegree sumDegree(const LaurentPolynomial<T>& p, 
		    const LaurentPolynomial<T>& q);
  template <class T>
  SDegree sumValuation(const LaurentPolynomial<T>& p, 
		       const LaurentPolynomial<T>& q);
};

/******** type definitions ***************************************************/

#include "vector.h"

namespace polynomials {
  using namespace vector;
};

namespace polynomials {

template <class T> class Polynomial {
 protected:
  Vector<T> v;
 public:
  typedef struct {} const_tag;
/* constructors and destructors */
  void operator delete(void* ptr)
    {return arena().free(ptr,sizeof(Polynomial<T>));}
  Polynomial<T>(){};
  Polynomial<T>(Degree d):v(d+1) {};
  Polynomial<T>(const Polynomial<T>& q):v(q.v) {};
  Polynomial<T>(T* const& ptr, const Degree& d):v(ptr,d+1) {};
  Polynomial<T>(const T& c, const_tag):v(1) {v[0] = c; setDegValue(0);}
  ~Polynomial<T>();
/* manipulators */
  T& operator[] (const Ulong& j) {return(v[j]);}
  T* ptr() {return v.ptr();}
  void reduceDeg() {v.reduceDim();}
  void setDeg(const Degree& d){v.setDim(d+1);}
  void setDegValue(const Degree& d){v.setDimValue(d+1);}
  void setVect(const T *source, Ulong n) {v.setVect(source,n);}
  void setZero(){v.dim() = 0;}
  void setZero(Ulong r){v.setZero(r);}
  void setZero(Ulong first, Ulong r){v.setZero(first,r);}
  Vector<T>& vect() {return(v);}
/* accessors */
  const T& operator[] (const Ulong& j) const {return(v[j]);}
  Ulong deg() const {return v.dim()-1;}
  bool isZero() const {return (deg() == undef_degree);}
  const T* ptr() const {return v.ptr();}
  const Vector<T>& vect() const {return(v);}
/* operators and operations */
  Polynomial<T>& operator= (const Polynomial<T>& q){v = q.v; return *this;}
  Polynomial<T>& operator+= (const Polynomial<T>& q);
  Polynomial<T>& operator-= (const Polynomial<T>& q);
  Polynomial<T>& operator*= (const T& a);
  Polynomial<T>& operator*= (const Polynomial<T>& q);
  Polynomial<T>& operator/= (const Polynomial<T>& q);
};

class Monomial
  {
  private:
    Degree n;
  public:
    Monomial(Degree d){n = d;};
  };

template <class T> class LaurentPolynomial {
 protected:
  Polynomial<T> d_pol;
  SDegree d_valuation; /* degree of first non-zero coefficient */
 public:
/* constructors and destructors */
  void operator delete(void* ptr)
    {return arena().free(ptr,sizeof(LaurentPolynomial<T>));}
  LaurentPolynomial<T>() {};
  LaurentPolynomial<T>(const SDegree& d, const SDegree& o = 0);
  ~LaurentPolynomial<T>();
/* accessors */
  const T& operator[] (const SDegree& j) const;                   /* inlined */
  
  bool operator== (const LaurentPolynomial& p) const;
  bool operator!= (const LaurentPolynomial& p) const;             /* inlined */
  bool operator<= (const LaurentPolynomial& p) const;
  bool operator>= (const LaurentPolynomial& p) const;
  bool operator< (const LaurentPolynomial& p) const;              /* inlined */
  bool operator> (const LaurentPolynomial& p) const;              /* inlined */

  SDegree deg() const;                                            /* inlined */
  bool isZero() const;                                            /* inlined */
  SDegree val() const;                                            /* inlined */

/* manipulators */
  T& operator[] (const SDegree& j);                               /* inlined */

  void adjustBounds();
  void setBounds(const SDegree& n, const SDegree& m);
  void setDeg(const SDegree& n);
  void setDegValue(const SDegree& n);                             /* inlined */
  void setVal(const SDegree& n);
  void setValValue(const SDegree& n);                             /* inlined */
  void setZero();                                                 /* inlined */
};

};

/******** inline definitions **************************************************/

namespace polynomials {

template <class T> 
inline bool operator!= (const Polynomial<T>& p, const Polynomial<T>& q)
  {return !(p == q);}
template <class T> 
inline bool operator< (const Polynomial<T>& p, const Polynomial<T>& q)
  {return !(p >= q);}
template <class T> 
inline bool operator> (const Polynomial<T>& p, const Polynomial<T>& q)
  {return !(p <= q);}

template<class T>
inline const T& LaurentPolynomial<T>::operator[] (const SDegree& j) const
  {return d_pol[j-d_valuation];}
template<class T> inline T& LaurentPolynomial<T>::operator[] (const SDegree& j)
  {return d_pol[j-d_valuation];}

template<class T>
inline bool LaurentPolynomial<T>::operator!= (const LaurentPolynomial& p)
  const {return !operator== (p);}
template<class T>
inline bool LaurentPolynomial<T>::operator> (const LaurentPolynomial& p)
  const {return !operator<= (p);}
template<class T>
inline bool LaurentPolynomial<T>::operator< (const LaurentPolynomial& p)
  const {return !operator>= (p);}

template<class T> inline SDegree LaurentPolynomial<T>::deg() const 
  {return d_pol.deg()+d_valuation;}
template<class T>
inline bool LaurentPolynomial<T>::isZero() const {return d_pol.isZero();}
template<class T> 
inline void LaurentPolynomial<T>::setDegValue(const SDegree& n)
  {d_pol.setDegValue(n-d_valuation);}
template<class T>
inline void LaurentPolynomial<T>::setValValue(const SDegree& n)
  {d_valuation = n;}
template<class T> inline SDegree LaurentPolynomial<T>::val() const 
  {return d_valuation;}
template<class T> inline void LaurentPolynomial<T>::setZero()
  {d_pol.setZero();}

};

#include "polynomials.hpp"

#endif
