/*
  This is special.cpp

  Coxeter version 3.0_demo  Copyright (C) 2001 Fokko du Cloux
  See file main.cpp for full copyright notice
*/

#include "special.h"

#include "directories.h"
#include "interactive.h"

#include "commands.h"
#include "fcoxgroup.h"
#include "cells.h"
#include "interactive.h"
#include "search.h"

#include <set>

/*****************************************************************************

  This module contains code to study the positivity of the structure constants 
  in the Hecke algebra, in the k-l c-basis. This asserts that if we write :

      c_x.c_y = sum_{z}b_{x,y,z}c_z

  then the coefficients b_{x,y,z} are (Laurent) polynomials with non-negative
  coefficients. It is known through deep geometrical arguments to hold when
  W is crystallographic.

  Our approach consists in computing c_x.c_y inductively, for fixed y, through
  the formula 

      c_x = c_s.c_{sx} - \sum_{z<sx,sz<z} mu(z,sx)c_z

  for sx < x, which then reduces the computation to things known recursively, 
  plus the action of c_s, which from the above formula is :

      c_s.c_v = c_sv + \sum_{u<v,su<u} m(u,v)c_u

  if sv > v, and if s.v < v, is given by c_s.c_v = (q^{1/2}+q^{-1/2}) c_v.

  So the basic input for this computation are the mu-coefficients for the
  group, which are readily accessible from the program.


  NOTE : the printouts always outputs only the nonnegative-degree part of
  the Laurent polynomials, as they are symmetric w.r.t. v -> v^{-1}.

******************************************************************************/

/******** local definitions **************************************************/

namespace {
  using namespace special;
  using namespace fcoxgroup;
  using namespace interactive;

  void special_f();
  void special_h();

  void cprod_f();
  void cycltable_f();
  void decrklpol_f();
  void klplist_f();
  void positivity_f();

  void cprod_h();
  void cycltable_h();
  void decrklpol_h();
  void klplist_h();
  void positivity_h();

  const char* special_tag = "user-defined command";

  const char* cprod_tag = "multiplies two elements in the k-l basis";
  const char* cycltable_tag = 
    "prints out the table of all c_x.c_y for a given y";
  const char* decrklpol_tag = "checks if k-l polynomials are decreasing";
  const char* klplist_tag = 
    "prints the list of all polynomials occurring as k-l polynomials";
  const char* positivity_tag = 
    "checks non-negativity of structure coefficients and writes them down";

  typedef unsigned Coeff;
  const Coeff COEFF_MAX = UINT_MAX;

  typedef Polynomial<Coeff> Pol;
  const Pol& one();

  // convenience classes

  class GraphData {
  private:
    List<List<List<CoxNbr> > > d_edge;
    List<List<List<KLCoeff> > > d_coeff;
    Ulong d_size;
  public:
    // constructors and destructors
    GraphData():d_edge(0),d_coeff(0),d_size(0) {};
    GraphData(const WGraph&, const Rank&);
    ~GraphData() {};
    // accessors
    const List<KLCoeff>& coeff(const Generator& s, const CoxNbr& x) const
      {return d_coeff[s][x];}
    const List<CoxNbr>& edge(const Generator& s, const CoxNbr& x) const
      {return d_edge[s][x];}
    const Ulong size() const {return d_size;}
  };

  // PolTree is a binary tree with additional data

  class PolTree:public search::BinaryTree<Pol> {
  private:
    Ulong d_maxCoeff;
  public:
    // constructors and destructors
    PolTree():d_maxCoeff(0) {}
    ~PolTree() {}
    // acessors
    Ulong maxCoeff() const {return d_maxCoeff;}
    // manipulators
    void reset();
    Pol* find(const Pol& p);
  };

  class PolTable {
  private:
    List<const Pol*> d_data;
    List<const Pol**> d_row;
    PolTree d_tree;
  public:
    // constructors and destructors
    PolTable() {};
    PolTable(const BitMap&);
    PolTable(const Ulong&, const Ulong&);
    ~PolTable() {};
    // accessors
    Ulong maxCoeff() const {return d_tree.maxCoeff();}
    const Pol** row(const Ulong& i) const {return d_row[i];}
    Ulong size() const {return d_row.size();}
    // manipulators
    const Pol*& operator()(const Ulong& i, const Ulong& j)
      {return d_row[i][j];}
    void setZero() {d_data.setZero();}
    void setZero(const Ulong& i, const Ulong& j) {d_row[i][j] = 0;}
    void write(const Pol& p, const Ulong& i, const Ulong& j)
      {d_row[i][j] = d_tree.find(p);}
    // temporary stuff
    PolTree& tree() {return d_tree;}
  };

  // functions for the positivity check

  void addSimpleTerm(List<Pol>&, const CoxNbr&, const Pol&);
  void addTerm(List<Pol>&, const List<CoxNbr>&, const List<KLCoeff>&, 
	       const Pol&);
  void extendTable(PolTable&, const WGraph&, const SchubertContext&, 
		   const CoxNbr&, const GraphData&);
  bool fillTable(PolTable&, const WGraph&, const SchubertContext&, 
		 const CoxNbr&, const GraphData&);
  bool isSmaller(const kl::KLPol&, const kl::KLPol&);
  bool isUnimodal(const Pol& pol);
  void prettyPrint(FILE*, const Pol& pol, const Degree&);
  void printRow(FILE*, const Pol**, FiniteCoxGroup*);
  void printTable(FILE*, const PolTable&, FiniteCoxGroup*);
  void safeAdd(Pol&, const Pol&);
  void safeAdd(Pol&, const Pol&, const KLCoeff&);
  void safeSubtract(Pol&, const Pol&);
  void safeSubtract(Pol&, const Pol&, const KLCoeff&);
  void subtract(List<Pol>&, const Pol**, const KLCoeff&);
  void writeRow(List<Pol>&, PolTable&, const Ulong& i);

};


/************************************************************************

        Chapter I -- Functions for the special commands.

  This section defines the functions provided by special.c

 ************************************************************************/


void special::addSpecialCommands(commands::CommandTree* tree)

/*
  This function should be edited if you want to add new commands to the
  system, instead of just editing the "special" command. Adding a command
  is easy : you should choose a name for it (an arbitrary string; if it
  already exists it will replace the existing command), a "tag" (used
  for help purposes; special_tag will do), a function which should take
  no arguments and return void, which will be executed by the command,
  and a help function (the empty function will do; this is already
  provided as relax_f, provided by commands.h).

  The commands are added to the main command tree of the program (the
  tree argument is used for convienience, since this function is called
  when mainCommandTree() is not yet functional); it would also be possible 
  to add special command trees, but we have decided to leave this to users 
  willing to delve into commands.c.

*/

{
  using namespace commands;

  /* prototype line for adding a special command */

  tree->add("special",special_tag,&special_f,&special_h);

  /* add user-defined commands here ... */

  tree->add("cprod",cprod_tag,&cprod_f,&cprod_h);
  tree->add("cycltable",cycltable_tag,&cycltable_f,&cycltable_h);
  tree->add("decrklpol",decrklpol_tag,&decrklpol_f,&decrklpol_h);
  tree->add("klplist",klplist_tag,&klplist_f,&klplist_h);
  tree->add("positivity",positivity_tag,&positivity_f,&positivity_h);

  return;
}

namespace {

void cprod_f()

/*
  Prints the product of two elements in the k-l basis. Tries to avoid
  unnecessary memory allocations, by allocating only rows in the table
  for elements <= x.
*/

{  
  using namespace commands;

  // check if the current group is finite; exit otherwise

  FiniteCoxGroup* W = dynamic_cast<FiniteCoxGroup*>(currentGroup());
  if (W == 0) {
    fprintf(stderr,"sorry, group must be finite\n");
    return;
  }

  W->fullContext(); // make schubert context available for the full group
                    // (this does not yet compute mu-coefficients)

  // get the two elements

  CoxWord g(0);

  printf("first : ");
  g = interactive::getCoxWord(W);
  if (ERRNO) {
    Error(ERRNO);
    return;
  }

  CoxNbr x = W->contextNumber(g);

  fprintf(stdout,"second : ");
  g = interactive::getCoxWord(W);
  if (ERRNO) {
    Error(ERRNO);
    return;
  }

  CoxNbr y = W->contextNumber(g);

  // fill the mu-table; make the W-graph

  W->fillMu();

  WGraph X(0);
  cells::lWGraph(X,W->kl());

  // initialize the computation

  GraphData gd(X,W->rank());

  BitMap bm(W->order());
  W->extractClosure(bm,x); // put in bm a bitmap representing [e,x]
  PolTable pt(bm);

  // initialize table for z = e

  pt.write(one(),0,y); // the others are initialized to 0 already

  // fill table

  BitMap::Iterator bm_end = bm.end();

  for (BitMap::Iterator i = ++bm.begin(); i != bm_end; ++i) {
    CoxNbr z = *i;
    extendTable(pt,X,W->schubert(),z,gd);
  }

  // print line

  printRow(stdout,pt.row(x),W);
  printf("\n");

  return;
}

void cycltable_f()

/*
  This command prompts the user for an element y in the group, then prints
  out the various c_x.c_y, for x in the group.
*/

{
  // check if the current group is finite; exit otherwise

  FiniteCoxGroup* W = dynamic_cast<FiniteCoxGroup*>(commands::currentGroup());
  if (W == 0) {
    fprintf(stderr,"sorry, group must be finite\n");
    return;
  }

  W->fullContext(); // make schubert context available for the full group
                    // (this does not yet compute mu-coefficients)

  // get the element

  CoxWord g(0);

  printf("enter your element (finish with a carriage return) :\n");
  g = interactive::getCoxWord(W);
  if (ERRNO) {
    Error(ERRNO);
    return;
  }

  // fill the mu-table; make the W-graph

  CoxNbr y = W->contextNumber(g);

  W->fillMu();

  WGraph X(0);
  cells::lWGraph(X,W->kl());

  // initialize the computation

  GraphData gd(X,W->rank());
  PolTable pt(W->order(),W->order());

  // fill table

  fillTable(pt,X,W->schubert(),y,gd);

  // output the result to a user-provided file

  OutputFile file; // the constructor lets the user specify a file
  printTable(file.f(),pt,W);

  return;
}

void decrklpol_f()

/*
  Checks whether for each x <= z <= y the polynomial P_{x,y}-P_{z,y} has
  non-negative coefficients. Here we may replace z by its extremalization,
  and then the same for x, so that it suffices to check this for x and
  z extremal.

  It is enough to do the check when y_inverse >= y.

  We proceed as follows : for each such y, we run through [e,y], and check
  that for all z in [e,y], and each coatom x of z, we have P_{z,y} <= P_{x,y}.
*/

{
  // check if the current group is finite; exit otherwise

  FiniteCoxGroup* W = dynamic_cast<FiniteCoxGroup*>(commands::currentGroup());
   if (W == 0) {
    fprintf(stderr,"sorry, group must be finite\n");
    return;
  }

  // set the context to the full group, and compute the k-l polynomials 

  W->fullContext();
  W->fillKL();

  // main loop

  for (CoxNbr y = 0; y < W->order(); ++y) {

    if (W->inverse(y) < y)
      continue;

    // traverse the interval [e,y]

    BitMap b(0);
    W->extractClosure(b,y); // put in b a bitmap representing [e,y]

    BitMap::Iterator b_end = b.end();

    for (BitMap::Iterator i = b.begin(); i != b_end; ++i) {

      CoxNbr z = *i; // current element of [e,y]
      const kl::KLPol& p_zy = W->klPol(z,y);

      // traverse the coatom set of z

      const CoatomList& c = W->coatoms(z);

      for (Ulong j = 0; j < c.size(); ++j) {

	CoxNbr x = c[j];
	const kl::KLPol& p_xy = W->klPol(x,y);

	if (!isSmaller(p_zy,p_xy)) { // found a counterexample
	                             // this is not expected to ever happen
	  fprintf(stderr,"error! x = %lu, z = %lu, y = %lu\n",x,y,z);
	  fprintf(stderr,"P_{x,y} = ");
	  print(stderr,p_xy,"q");
	  fprintf(stderr,"\n");
	  fprintf(stderr,"P_{z,y} = ");
	  print(stderr,p_zy,"q");
	  fprintf(stderr,"\n");
	  return;
	}
      }
    }
    printf("%lu\r",y); // print a counter to show progress
  }

  printf("check was successful\n");

  return;
}

void klplist_f()

/*
  Prints the full list of all distinct k-l polynomials.
*/

{
  // check if the current group is finite; exit otherwise

  FiniteCoxGroup* W = dynamic_cast<FiniteCoxGroup*>(commands::currentGroup());
  if (W == 0) {
    fprintf(stderr,"sorry, group must be finite\n");
    return;
  }

  // set the context to the full group, and compute the k-l polynomials

  W->fullContext();
  W->fillKL();

  // output the list of polynomials to a user-defined file

  OutputFile file; // the constructor lets the user specify a file
  print(file.f(),W->kl().tree().root(),"q");

  return;
}

void positivity_f()

/*
  Checks out the non-negativity of the structure constants of the Hecke
  algebra in the k-l basis.

  The function starts out by checking that the group is finite. Then it
  makes sure that all the mu-coefficients for W are available, and initializes
  the tables that are used in the computation : gd holds the data from the
  W-graph in the precise form in which it shall be needed, and pt is the
  main table for the computation, consisting of a matrix of |W|^2 polynomial
  pointers, and a tree to hold the actual polynomials. For H4, the table of 
  polynomial pointers is the biggest memory requirement, and takes up about 
  800 MB for a 32-bit system (and twice that amount on a 64-bit system!)

  The computation then proceeds as explained in the article, filling the
  table for fixed y. The only material memory-requirement from now on comes
  from the growth of pt.d_tree. Here we do not try to keep the polynomials
  so we simply flush the tree at each pass.
*/

{
  // check if the current group is finite; exit otherwise

  FiniteCoxGroup* W = dynamic_cast<FiniteCoxGroup*>(commands::currentGroup());
  if (W == 0) {
    fprintf(stderr,"sorry, group must be finite\n");
    return;
  }

  // set the context to the full group, and compute the mu coefficients

  W->fullContext();
  W->fillMu();

  // construct the (left) W-graph of the full group

  WGraph X(0);
  cells::lWGraph(X,W->kl());

  // initialize the computation

  GraphData gd(X,W->rank());
  PolTable pt(W->order(),W->order());

  // reset positivity_log and error_log

  fclose(fopen("positivity_log","w"));
  fclose(fopen("error_log","w"));

  // run through computation

  for (Ulong y = 0; y < W->order(); ++y) {
    pt.setZero();
    fillTable(pt,X,W->schubert(),y,gd);
    pt.tree().reset();
    FILE* log = fopen("positivity_log","a");
    fprintf(log,"%lu: maxcoeff = %lu\n",y,pt.maxCoeff());
    fclose(log);
  }

  return;
}

void special_f()

{
  fprintf(stderr,"not implemented\n");
  return;
}

void cprod_h()

{
  printFile(stderr,"cprod.help",directories::MESSAGE_DIR);
  return;
}

void cycltable_h()

{
  printFile(stderr,"cycltable.help",directories::MESSAGE_DIR);
  return;
}

void decrklpol_h()

{
  printFile(stderr,"decrklpol.help",directories::MESSAGE_DIR);
  return;
}

void klplist_h()

{
  printFile(stderr,"klplist.help",directories::MESSAGE_DIR);
  return;
}

void positivity_h()

{
  printFile(stderr,"positivity.help",directories::MESSAGE_DIR);
  return;
}

void special_h()

{}

};

/*****************************************************************************

        Chapter II -- The GraphData class.

  This class makes available the data from the W-graph of the group which are
  required in the computation, in a convenient way. The idea is that d_edge
  lists for each generator s, and each z in the group s.t. sz < z, the set
  of x s.t. (z,x) is an edge in the graph, and sx > x; these are the terms
  that occur in the formulas for c_s.c_x. The list d_coeff contains the
  corresponding mu-coefficient.

  The point is to do these extractions once and for all, and record them here,
  instead of having to run zillions of times through the edge-lists of the
  graph; the memory cost is very reasonable.

******************************************************************************/

namespace {

GraphData::GraphData(const WGraph& X, const Rank& l)

/*
  Constructs the edge and coefficient lists for each generator, as explained
  in the introduction to this section.
*/

{  
  d_size = X.size();

  // allocate lists

  d_edge.setSize(l);
  d_coeff.setSize(l);

  for (Generator s = 0; s < l; ++s) {
    (d_edge)[s].setSize(d_size);
    (d_coeff)[s].setSize(d_size);
  }

  // fill lists

  for (CoxNbr x = 0; x < X.size(); ++x) {
    const EdgeList& e = X.edge(x);
    const CoeffList& c = X.coeffList(x);
    LFlags fx = X.descent(x);
    for (Ulong j = 0; j < e.size(); ++j) {
      CoxNbr z = e[j];
      LFlags fz = X.descent(z);
      // look at the generators s.t. sx < x and sz > z
      for (LFlags f = fx & ~fz; f; f &= f-1) {
	Generator s = firstBit(f);
	d_edge[s][z].append(x);
	d_coeff[s][z].append(c[j]);
      }
    }
  }
}

};

/*****************************************************************************

        Chapter III -- The PolTable class.

  This is another helper class for the computation, containing the data for
  the management of the polynomial table. This is a huge matrix of polynomial
  pointers, one for each pair of elements in the group, contained in d_list;
  the actual polynomials are stored in d_tree, a binary tree which makes it
  easy to search for them.

  One important fact is that zero polynomials are ignored, i.e. correspond
  to "empty" entries in the table (zero-pointers).

******************************************************************************/

namespace {

PolTable::PolTable (const Ulong& m, const Ulong& n)

/*
  Allocates an empty (i.e., zero-pointer initialized) table of size size.size.
  The idea is to allocate the memory in a single large block, and let each
  row point to the appropriate offset.

  This is done primarily to optimize the setZero member function. Believe it
  or not, setting each individual row to zero made this into the dominant
  function of the program!
*/

{
  d_data.setSize(m*n);
  d_row.setSize(m);

  d_row[0] = &d_data[0];

  for (Ulong j = 1; j < m; ++j)
    d_row[j] = d_row[j-1] + n;
}

PolTable::PolTable (const BitMap& b)

/*
  Does a "sparse" allocation of the table, allocating rows only for the
  elements flagged by the bitmap. This is used for the computation of a
  single product c_x.c_y; only products c_u.c_y for u <= x are required
  in the inductive computation.
*/

{
  Ulong m = b.bitCount();
  Ulong n = b.size();

  d_data.setSize(m*n);
  d_row.setSize(n);

  BitMap::Iterator b_end = b.end();

  Ulong j = 0;

  for (BitMap::Iterator i = b.begin(); i != b_end; ++i, ++j)
    d_row[*i] = &d_data[j*n];
}

};

/*****************************************************************************

        Chapter IV -- The PolTree class

  This is a small modification of a binary tree, where the size of the 
  allocations is recorded as we go.

******************************************************************************/

namespace {

Pol* PolTree::find(const Pol& p)

/*
  Looks up a polynomial in the tree, adding it if it is not already there.

*/

{
  Ulong prev_size = size();
  Pol* r = search::BinaryTree<Pol>::find(p);

  if (size() > prev_size) { // new polynomial found

    // check unimodality
    
    if (!isUnimodal(p)) { // polynomial is not unimodal!
                          // write error message
      FILE* errorFile = fopen("error_log","a");
      fprintf(errorFile,"\nnon-unimodal polynomial found :\n\n");
      fclose(errorFile);
    }

    // update maxCoeff

    for (Ulong i = 0; i <= p.deg(); ++i) {
      if (p[i] > d_maxCoeff)
	d_maxCoeff = p[i];
    }
  }

  return r;
}

void PolTree::reset()

/*
  Frees the memory for the tree part, while retaining the maxCoeff information.
*/

{
  delete d_root;
  d_root = 0;
  d_size = 0;
}

};

/*****************************************************************************

        Chapter V -- Functions for the positivity check.

******************************************************************************/

namespace {

void addSimpleTerm(List<Pol>& a, const CoxNbr& z, const Pol& q)

/*
  Deals with the case where we have to add (1/v+v)*q to a. In fact, we write
  only the positive part of the polynomial; this has either all even or all
  odd terms. Then the effect of multiplying by (1/v+v) is as follows :
  (a) degree increases by one (b) if j > 1, the j-th coefficient of the new
  polynomial is the sum of the two surrounding coefficients in the old one;
  (c) when the degree is even, the 0-th coefficient of the new polynomial
  is twice the 1-st coefficient of the old one.

  NOTE : we must be very careful to check for overflow. It so happens that
  there is in fact no overflow on a 32-bit computer for H4, but it's not
  so far off.
*/

{
  Pol& p = a[z];

  if (p.isZero() || (p.deg() < q.deg()+1))
    p.setDeg(q.deg()+1);

  if (q.deg()%2) { // odd degree for q
    if (p[0] > (COEFF_MAX-2*q[1]))
      goto overflow;
    p[0] += 2*q[1];
    for (Ulong j = 2; j < q.deg(); j += 2) {
      if (p[j] > (COEFF_MAX-q[j-1]))
	goto overflow;
      p[j] += q[j-1];
      if (p[j] > (COEFF_MAX-q[j+1]))
	goto overflow;
      p[j] += q[j+1];
    }
    if (p[q.deg()+1] > (COEFF_MAX-q[q.deg()]))
      goto overflow;
    p[q.deg()+1] += q[q.deg()];
  }
  else { // even degree for q
    for (Ulong j = 1; j < q.deg(); j += 2) {
      if (p[j] > (COEFF_MAX-q[j-1]))
	goto overflow;
      p[j] += q[j-1];
      if (p[j] > (COEFF_MAX-q[j+1]))
	goto overflow;
      p[j] += q[j+1];
    }
    if (p[q.deg()+1] > (COEFF_MAX-q[q.deg()]))
      goto overflow;
    p[q.deg()+1] += q[q.deg()];
  }

  return;

 overflow:

  FILE* errorFile = fopen("error_log","a");
  fprintf(errorFile,"\noverflow in addSimpleTerm\n\n");
  fclose(errorFile);

  return;
}

void addTerm(List<Pol>& a, const List<CoxNbr>& e, const List<KLCoeff>& c, 
	     const Pol& pol)

/*
  Adds to a, representing the coefficients of an element of the Hecke
  algebra in the c-basis, pol times the element represented in sparse
  form by e and c.
*/

{
  for (Ulong j = 0; j < e.size(); ++j) {

    if (c[j] == 1) // this happens allmost all the time
                   // call an optimized version
      safeAdd(a[e[j]],pol);
    else {
      safeAdd(a[e[j]],pol,c[j]);
    }
  }

  return;
}

void extendTable(PolTable& pt, const WGraph& X, const SchubertContext& p, 
		 const CoxNbr& x, const GraphData& gd)

/*
  Fills in the row for x in b, assuming that previous rows have been filled
  already. Uses the formula
*/

{
  static List<Pol> a(0);
  a.setSize(gd.size());

  LFlags f = X.descent(x);    // f holds the left descent set of x
  Generator s = lastBit(f);   // largest descent generator
#if 0
  Generator s = firstBit(f);  // smallest descent generator
#endif
  CoxNbr sx = p.lshift(x,s);

  // compute c_s.c_sx.c_y

  for (CoxNbr z = 0; z < gd.size(); ++z) {

    const Pol* pol_ptr = pt(sx,z); // this is b_{sx,y,z}
    if (pol_ptr == 0)
      continue;

    if (X.descent(z) & lmask[s])  // a[z] += (v+v^{-1})b_{sx,y,z}
      addSimpleTerm(a,z,*pol_ptr);

    else { // c_s.c_z is computed from the W-graph using the data in gd

      const List<CoxNbr>& e = gd.edge(s,z);
      const List<KLCoeff>& c = gd.coeff(s,z);
      addTerm(a,e,c,*pol_ptr);
    }
  }

  // subtract terms mu(z,sx)c_z.c_y

  const List<CoxNbr>& e = gd.edge(s,sx);
  const List<KLCoeff>& c = gd.coeff(s,sx);

  for (Ulong j = 0; j < e.size(); ++j) {
    CoxNbr z = e[j];
    if (z == x)
      continue;
    subtract(a,pt.row(z),c[j]);
  }

  // write row to table

  writeRow(a,pt,x);

  return;
}

bool fillTable(PolTable& pt, const WGraph& X, const SchubertContext& p, 
	       const CoxNbr& y, const GraphData& gd)

/*
  This function fills in the table of coefficients b_{x,y,z} for fixed y,
  i.e., it computes the various c_x.c_y when x runs through the group.
  It works by induction on the length of x, by successive calls to
  extendTable (see the introduction to this module for more details.)
*/

{  
  // initialize row 0

  pt.write(one(),0,y); // the others are initialized to 0 already

  // build table

  for (CoxNbr x = 1; x < gd.size(); ++x)
    extendTable(pt,X,p,x,gd);

  return true;
}

bool isSmaller(const kl::KLPol& p, const kl::KLPol& q)

/*
  Returns true if q-p has nonnegative coefficients; false otherwise. It is 
  assumed that p and q are non-zero.
*/

{
  if (p.deg() > q.deg())
    return false;

  for (Ulong j = 0; j <= p.deg(); ++j) {
    if (p[j] > q[j])
      return false;
  }

  return true;
}

bool isUnimodal(const Pol& pol)

/*
  Tells if the polynomial pol is unimodal (in fact, we are passed the second
  half of the polynomial; so the check here is to see if p[j] <= p[j-1] for
  all j >= 1.)
*/

{
  for (Degree j = pol.deg()%2 + 2; j <= pol.deg(); j += 2)
    if (pol[j] > pol[j-2])
      return false;

  return true;
}

const Pol& one()

/*
  Returns the constant polynomial equal to one (this makes for simpler code
  elsewhere.)
*/

{
  static Pol pol(1,Pol::const_tag());
  return pol;
}

void prettyPrint(FILE* file, const Pol& pol, const Degree& parity)

/*
  Pretty-prints the polynomial onto file, with the appropriate degree-expansion
  according to parity.
*/

{
  static Pol p(0);

  p.setZero();
  p.setDeg(2*pol.deg()+parity);

  for (Degree j = 0; j <= pol.deg(); ++j)
    p[2*j+parity] = pol[j];

  print(file,p,"v");

  return;
}

void printRow(FILE* file, const Pol** a, FiniteCoxGroup* W)

/*
  Prints out a row in the coefficient table. This corresponds to the expression
  of c_x.c_y for some x and y.

  NOTE: as always, only the "positive-degree-half" of the polynomials are
  actually printed; the other half is symmetric.
*/

{  
  bool first = true; // tells if we are printing the first non-zero coefficient

  for (Ulong i = 0; i < W->order(); ++i) {

    if (a[i] == 0)
      continue;

    // print a "+" only if this is not the first term

    if (first)
      first = false;
    else
      fprintf(file,"+");

    // print output in the form (polynomial_in_v)c(group_element)

    fprintf(file,"(");
    print(file,*(a[i]),"v");
    fprintf(file,")c(",i);
    W->print(file,i);
    fprintf(file,")",i);
  }
  
  return;
}

void printTable(FILE* file, const PolTable& pt, FiniteCoxGroup* W)

/*
  Prints the full table of polynomials, for a given y, by successive calls to
  printRow.
*/

{
  for (Ulong j = 0; j < pt.size(); ++j) {
    bool first = true;
    const Pol** bpp = pt.row(j);
    W->print(file,j);
    fprintf(file,":");
    printRow(file,bpp,W);
    fprintf(file,"\n");
  }

  return;
}

void safeAdd(Pol& p, const Pol& q)

/*
   Adds q to p while looking out for overflow. Makes a note in error_log
   if overflow is seen.
*/

{
  if (p.isZero()) {
    p = q;
    return;
  }

  if (p.deg() < q.deg())
    p.setDeg(q.deg());

  for (Ulong j = 0; j <= q.deg(); ++j) {
    if (p[j] > (COEFF_MAX-q[j]))
      goto overflow;
    p[j] += q[j];
  }

  return;

 overflow:

  FILE* errorFile = fopen("error_log","a");
  fprintf(errorFile,"\noverflow in addTerm\n\n");
  fclose(errorFile);

  return;
}

void safeAdd(Pol& p, const Pol& q, const KLCoeff& c)

/*
   Adds c*q to p while looking out for overflow. Makes a note in error_log
   if overflow is seen.
*/

{
  if (p.isZero() || (p.deg() < q.deg()))
    p.setDeg(q.deg());

  for (Ulong j = 0; j <= q.deg(); ++j) {
    if (q[j] > COEFF_MAX/c)
      goto overflow;
    if (p[j] > (COEFF_MAX-c*q[j]))
      goto overflow;
    p[j] += c*q[j];
  }

  return;

 overflow:

  FILE* errorFile = fopen("error_log","a");
  fprintf(errorFile,"\noverflow in addTerm\n\n");
  fclose(errorFile);

  return;
}

void safeSubtract(Pol& p, const Pol& q)

/*
   Subtracts q from p while looking out for overflow. Makes a note in error_log
   if underflow is seen.

   NOTE : it is assumed that q is non-zero.
*/

{
  if (p.isZero() || (p.deg() < q.deg())) {
    print(stderr,p,"v");
    printf("\n");
    print(stderr,q,"v");
    printf("\n");
    goto underflow;
  }

  for (Ulong j = 0; j <= q.deg(); ++j) {
    if (p[j] < q[j])
      goto underflow;
    p[j] -= q[j];
  }

  p.reduceDeg();

  return;

 underflow:

  FILE* errorFile = fopen("error_log","a");
  fprintf(errorFile,"\nunderflow in subtract\n\n");
  fclose(errorFile);

  return;
}

void safeSubtract(Pol& p, const Pol& q, const KLCoeff& c)

/*
   Subtracts c*q to p while looking out for underflow. Makes a note in 
   error_log if underflow is seen.

   NOTE : it is assumed that q is non-zero!
*/

{
  if (p.isZero() || (p.deg() < q.deg()))
    goto underflow;

  for (Ulong j = 0; j <= q.deg(); ++j) {
    if (q[j] > COEFF_MAX/c)
      goto underflow;
    if (p[j] < c*q[j])
      goto underflow;
    p[j] -= c*q[j];
  }

  p.reduceDeg();

  return;

 underflow:

  FILE* errorFile = fopen("error_log","a");
  fprintf(errorFile,"\nunderflow in addTerm\n\n");
  fclose(errorFile);

  return;
}

void subtract(List<Pol>& a, const Pol** b, const KLCoeff& c)

/*
  Subtracts b*c from a;
*/

{
  if (c > 1) {
    for (Ulong j = 0; j < a.size(); ++j) {
      if (b[j] == 0)
	continue;
      safeSubtract(a[j],*(b[j]),c);
    }
  }
  else {
    for (Ulong j = 0; j < a.size(); ++j) {
      if (b[j] == 0)
	continue;
      safeSubtract(a[j],*(b[j]));
    }
  }

  return;
}

void writeRow(List<Pol>& a, PolTable& pt, const Ulong& i)

/*
  Writes down the row a in the table. It sets the polynomials in a to zero
  as it reads them.
*/

{  
  for (Ulong j = 0; j < a.size(); ++j) {
    if (a[j].isZero())
      continue;
    pt.write(a[j],i,j);
    a[j].setZero();
  }

  return;
}

};

